# frozen_string_literal: true

# Released under the MIT License.
# Copyright, 2013-2019, by Sam Saffron.
# Copyright, 2015-2016, by Dave Gynn.
# Copyright, 2018, by Jonas Peschla.
# Copyright, 2020-2025, by Samuel Williams.

module Memory
	VERSION = "0.12.0"
end
